@@ -26,8 +26,10 @@
 #include <internal/filesystem.hpp>
 #include <internal/debdeltahelper.hpp>
 #include <internal/lock.hpp>
+#include <internal/cachefiles.hpp>
 
 #include <internal/worker/packages.hpp>
+#include <internal/worker/dpkg.hpp>
 
 namespace cupt {
 namespace internal {
@@ -63,6 +65,18 @@ string InnerAction::toString() const
 	return result;
 }
 
+InnerAction::Type InnerActionGroup::getCompoundActionType() const
+{
+	/* all actions within one group have, by algorithm,
+	   a) the same action name (so far, configures only)
+	   b) the same package name (linked actions)
+
+	   in both cases, we can choose the action type of the last action
+	   in the subgroup as effective
+	*/
+	return this->rbegin()->type;
+}
+
 
 // Attribute
 GraphAndAttributes::Attribute::Attribute()
@@ -123,7 +137,7 @@ PackagesWorker::PackagesWorker()
 set< string > __get_pseudo_essential_package_names(const Cache& cache, bool debugging)
 {
 	set< string > result;
-	queue< shared_ptr< const BinaryVersion > > toProcess;
+	queue< const BinaryVersion* > toProcess;
 
 	auto processRelationExpression = [&cache, &result, &toProcess, &debugging](const RelationExpression& relationExpression)
 	{
@@ -179,6 +193,20 @@ set< string > __get_pseudo_essential_package_names(const Cache& cache, bool debu
 	return result;
 }
 
+const BinaryVersion* PackagesWorker::__get_fake_version_for_purge(const string& packageName)
+{
+	auto& versionPtr = __fake_versions_for_purge[packageName];
+	if (!versionPtr)
+	{
+		versionPtr.reset(new BinaryVersion);
+		versionPtr->packageName = packageName;
+		versionPtr->versionString = "<dummy>";
+		versionPtr->essential = false;
+	}
+
+	return versionPtr.get();
+}
+
 void __set_action_priority(const InnerAction* actionPtr, const InnerAction* previousActionPtr)
 {
 	/* priorities are assigned that way so the possible chains are sorted in
@@ -223,6 +251,7 @@ void PackagesWorker::__fill_actions(GraphAndAttributes& gaa)
 		{ Action::Install, vector< IA::Type >{ IA::Unpack, IA::Configure } },
 		{ Action::Upgrade, vector< IA::Type >{ IA::Remove, IA::Unpack, IA::Configure } },
 		{ Action::Downgrade, vector< IA::Type >{ IA::Remove, IA::Unpack, IA::Configure } },
+		{ Action::Reinstall, vector< IA::Type >{ IA::Remove, IA::Unpack, IA::Configure } },
 		{ Action::Configure, vector< IA::Type >{ IA::Configure } },
 		{ Action::Deconfigure, vector< IA::Type >{ IA::Remove } },
 		{ Action::Remove, vector< IA::Type >{ IA::Remove } },
@@ -258,7 +287,7 @@ void PackagesWorker::__fill_actions(GraphAndAttributes& gaa)
 			{
 				const IA::Type& innerActionType = *innerActionTypeIt;
 
-				shared_ptr< const BinaryVersion > version;
+				const BinaryVersion* version = nullptr;
 				if (innerActionType == IA::Remove)
 				{
 					auto package = _cache->getBinaryPackage(packageName);
@@ -273,11 +302,7 @@ void PackagesWorker::__fill_actions(GraphAndAttributes& gaa)
 				}
 				if (!version)
 				{
-					auto versionPtr = new BinaryVersion;
-					versionPtr->packageName = packageName;
-					versionPtr->versionString = "<dummy>";
-					versionPtr->essential = false;
-					version.reset(versionPtr);
+					version = __get_fake_version_for_purge(packageName);
 				}
 
 				if (innerActionType != IA::Unpack)
@@ -339,6 +364,13 @@ void __fill_action_dependencies(FillActionGeneralInfo& gi,
 		BinaryVersion::RelationTypes::Type dependencyType, InnerAction::Type actionType,
 		Direction::Type direction)
 {
+	typedef BinaryVersion::RelationTypes RT;
+	if (gi.innerActionPtr->fake &&
+			(dependencyType != RT::PreDepends && dependencyType != RT::Depends))
+	{
+		return;
+	}
+
 	const set< InnerAction >& verticesMap = gi.gaaPtr->graph.getVertices();
 
 	InnerAction candidateAction;
@@ -481,23 +513,9 @@ void __fill_graph_dependencies(const shared_ptr< const Cache >& cache,
 	}
 }
 
-const shared_ptr< const BinaryVersion > __create_virtual_version(
-		const shared_ptr< const BinaryVersion >& version)
-{
-	typedef BinaryVersion::RelationTypes RT;
-
-	shared_ptr< BinaryVersion > virtualVersion(new BinaryVersion);
-	virtualVersion->packageName = version->packageName;
-	virtualVersion->versionString = version->versionString;
-	virtualVersion->relations[RT::PreDepends] = version->relations[RT::PreDepends];
-	virtualVersion->relations[RT::Depends] = version->relations[RT::Depends];
-	virtualVersion->essential = false;
-	return virtualVersion;
-}
-
 void __create_virtual_edge(
-		const shared_ptr< const BinaryVersion >& fromVirtualVersion,
-		const shared_ptr< const BinaryVersion >& toVirtualVersion,
+		const BinaryVersion* fromVirtualVersion,
+		const BinaryVersion* toVirtualVersion,
 		vector< pair< InnerAction, InnerAction > >* virtualEdgesPtr)
 {
 	InnerAction from;
@@ -531,11 +549,6 @@ vector< pair< InnerAction, InnerAction > > __create_virtual_actions(
 	   which goes into
 
 	   'install y' -> 'remove x'
-
-	   moreover, we split the installed version into virtual version by one
-	   relation expression, so different relation expressions of the same real
-	   version don't interact with each other, otherwise we'd get a full cyclic
-	   mess
 	*/
 	vector< pair< InnerAction, InnerAction > > virtualEdges;
 
@@ -555,24 +568,23 @@ vector< pair< InnerAction, InnerAction > > __create_virtual_actions(
 			continue;
 		}
 
-		auto virtualVersion = __create_virtual_version(installedVersion);
-		__create_virtual_edge(virtualVersion, virtualVersion, &virtualEdges);
+		__create_virtual_edge(installedVersion, installedVersion, &virtualEdges);
 	}
 
 	return virtualEdges;
 }
 
-bool __share_relation_expression(const GraphAndAttributes::Attribute& left,
-		const GraphAndAttributes::Attribute& right)
+const GraphAndAttributes::RelationInfoRecord* __get_shared_relation_info_record(
+		const GraphAndAttributes::Attribute& left, const GraphAndAttributes::Attribute& right)
 {
-	FORIT(leftRelationRecordIt, left.relationInfo)
+	for (const auto& leftRelationRecord: left.relationInfo)
 	{
-		FORIT(rightRelationRecordIt, right.relationInfo)
+		for (const auto& rightRelationRecord: right.relationInfo)
 		{
-			if (leftRelationRecordIt->dependencyType == rightRelationRecordIt->dependencyType &&
-					leftRelationRecordIt->relationExpression == rightRelationRecordIt->relationExpression)
+			if (leftRelationRecord.dependencyType == rightRelationRecord.dependencyType &&
+					leftRelationRecord.relationExpression == rightRelationRecord.relationExpression)
 			{
-				return true;
+				return &leftRelationRecord;
 			}
 		}
 	}
@@ -657,35 +669,37 @@ void __expand_and_delete_virtual_edges(GraphAndAttributes& gaa,
 		const InnerAction* toPtr = gaa.graph.addVertex(edgeIt->second);
 
 		// "multiplying" the dependencies
-		const GraphCessorListType predecessors = gaa.graph.getPredecessorsFromPointer(fromPtr);
-		const GraphCessorListType successors = gaa.graph.getSuccessorsFromPointer(toPtr);
-		FORIT(predecessorVertexPtrIt, predecessors)
+		const auto& predecessors = gaa.graph.getPredecessorsFromPointer(fromPtr);
+		const auto& successors = gaa.graph.getSuccessorsFromPointer(toPtr);
+		for (auto predecessor: predecessors)
 		{
-			FORIT(successorVertexPtrIt, successors)
+			for (auto successor: successors)
 			{
-				if (*predecessorVertexPtrIt == *successorVertexPtrIt)
-				{
-					continue;
-				}
-				if (!__share_relation_expression(
-							gaa.attributes[make_pair(*predecessorVertexPtrIt, fromPtr)],
-							gaa.attributes[make_pair(toPtr, *successorVertexPtrIt)]))
-				{
-					continue;
-				}
+				if (predecessor == successor) continue;
 
-				__move_edge(gaa, *predecessorVertexPtrIt, fromPtr, *predecessorVertexPtrIt, *successorVertexPtrIt, debugging);
-				__move_edge(gaa, toPtr, *successorVertexPtrIt, *predecessorVertexPtrIt, *successorVertexPtrIt, debugging);
+				auto sharedRir = __get_shared_relation_info_record(
+						gaa.attributes[make_pair(predecessor, fromPtr)],
+						gaa.attributes[make_pair(toPtr, successor)]);
+				if (!sharedRir) continue;
+
+				gaa.graph.addEdgeFromPointers(predecessor, successor);
+				gaa.attributes[make_pair(predecessor, successor)].relationInfo.push_back(*sharedRir);
 				if (debugging)
 				{
 					const string& mediatorPackageName = fromPtr->version->packageName;
 					debug2("multiplied action dependency: '%s' -> '%s', virtual mediator: '%s'",
-							(*predecessorVertexPtrIt)->toString(), (*successorVertexPtrIt)->toString(), mediatorPackageName);
+							predecessor->toString(), successor->toString(), mediatorPackageName);
 				}
-
 			}
 		}
-
+		for (auto predecessor: predecessors)
+		{
+			gaa.attributes.erase(make_pair(predecessor, fromPtr));
+		}
+		for (auto successor: successors)
+		{
+			gaa.attributes.erase(make_pair(toPtr, successor));
+		}
 		gaa.graph.deleteVertex(*fromPtr);
 		gaa.graph.deleteVertex(*toPtr);
 	}
@@ -710,9 +724,8 @@ void __expand_linked_actions(const Cache& cache, GraphAndAttributes& gaa, bool d
 			if (relationRecordIt->reverse == neededValueOfReverse)
 			{
 				auto satisfyingVersions = cache.getSatisfyingVersions(relationRecordIt->relationExpression);
-				auto predicate = std::bind2nd(PointerEqual< const BinaryVersion >(), antagonisticPtr->version);
-				if (std::find_if(satisfyingVersions.begin(), satisfyingVersions.end(),
-						predicate) == satisfyingVersions.end())
+				if (std::find(satisfyingVersions.begin(), satisfyingVersions.end(),
+						antagonisticPtr->version) == satisfyingVersions.end())
 				{
 					return false;
 				}
@@ -1295,24 +1308,9 @@ void __split_heterogeneous_actions(const shared_ptr< const Cache >& cache, Logge
 	newActionGroups.swap(actionGroups);
 }
 
-static string __get_codename_and_component_string(const Version& version, const string& baseUri)
-{
-	vector< string > parts;
-	FORIT(sourceIt, version.sources)
-	{
-		auto releaseInfo = sourceIt->release;
-		if (releaseInfo->baseUri != baseUri)
-		{
-			continue;
-		}
-		parts.push_back(releaseInfo->codename + '/' + releaseInfo->component);
-	}
-	return join(",", parts);
-}
-
 static string __get_long_alias_tail(const Version& version, const string& baseUri)
 {
-	return format2("%s %s %s", __get_codename_and_component_string(version, baseUri),
+	return format2("%s %s %s", version.getCodenameAndComponentString(baseUri),
 			version.packageName, version.versionString);
 }
 
@@ -1343,13 +1341,12 @@ map< string, pair< download::Manager::DownloadEntity, string > > PackagesWorker:
 
 		DebdeltaHelper debdeltaHelper;
 
-		static const vector< Action::Type > actions = { Action::Install, Action::Upgrade, Action::Downgrade };
-		FORIT(actionIt, actions)
+		for (auto actionType: _download_dependent_action_types)
 		{
-			const Resolver::SuggestedPackages& suggestedPackages = __actions_preview->groups[*actionIt];
+			const auto& suggestedPackages = __actions_preview->groups[actionType];
 			FORIT(it, suggestedPackages)
 			{
-				const shared_ptr< const BinaryVersion >& version = it->second.version;
+				const auto& version = it->second.version;
 
 				const string& packageName = version->packageName;
 				const string& versionString = version->versionString;
@@ -1754,7 +1751,7 @@ string PackagesWorker::__generate_input_for_preinstall_v2_hooks(
 		FORIT(actionIt, *actionGroupIt)
 		{
 			auto actionType = actionIt->type;
-			const shared_ptr< const BinaryVersion >& version = actionIt->version;
+			const auto& version = actionIt->version;
 			string path;
 			switch (actionType)
 			{
@@ -1935,7 +1932,7 @@ void PackagesWorker::markAsAutomaticallyInstalled(const string& packageName, boo
 			{
 				__auto_installed_package_names.erase(packageName);
 			}
-			auto extendedInfoPath = _cache->getPathOfExtendedStates();
+			auto extendedInfoPath = cachefiles::getPathOfExtendedStates(*_config);
 			fs::mkpath(fs::dirname(extendedInfoPath));
 			auto tempPath = extendedInfoPath + ".cupt.tmp";
 
@@ -1983,13 +1980,6 @@ void PackagesWorker::__do_downloads(const vector< pair< download::Manager::Downl
 		{
 			Lock lock(*_config, archivesDirectory + "/lock");
 
-			uint64_t totalDownloadSize = 0;
-			FORIT(it, downloads)
-			{
-				totalDownloadSize += it->first.size;
-			}
-			downloadProgress->setTotalEstimatedSize(totalDownloadSize);
-
 			vector< download::Manager::DownloadEntity > params;
 			FORIT(it, downloads)
 			{
@@ -2009,41 +1999,27 @@ void PackagesWorker::__do_downloads(const vector< pair< download::Manager::Downl
 	}
 }
 
-string __get_dpkg_action_log(const InnerActionGroup& actionGroup,
-		InnerAction::Type actionType, const string& actionName)
+void PackagesWorker::__do_independent_auto_status_changes()
 {
-	vector< string > subResults;
-	FORIT(actionIt, actionGroup)
+	auto wasDoneAlready = [this](const string& packageName)
 	{
-		if (actionIt->type == actionType)
+		for (const auto& actionGroup: __actions_preview->groups)
 		{
-			subResults.push_back(actionName + ' ' + actionIt->version->packageName
-					+ ' ' + actionIt->version->versionString);
+			if (actionGroup.count(packageName))
+			{
+				return true;
+			}
 		}
-	}
-	return join(" & ", subResults);
-}
+		return false;
+	};
 
-bool __defer_triggers(const Config& config, const Cache& cache)
-{
-	const string& optionName = "cupt::worker::defer-triggers";
-	if (config.getString(optionName) == "auto")
+	for (const auto& autoFlagChange: __actions_preview->autoFlagChanges)
 	{
-		auto dpkgPackage = cache.getBinaryPackage("dpkg");
-		if (!dpkgPackage)
+		const auto& packageName = autoFlagChange.first;
+		if (!wasDoneAlready(packageName))
 		{
-			fatal2i("no 'dpkg' binary package available");
+			markAsAutomaticallyInstalled(packageName, autoFlagChange.second);
 		}
-		auto dpkgInstalledVersion = dpkgPackage->getInstalledVersion();
-		if (!dpkgInstalledVersion)
-		{
-			fatal2i("no installed version for 'dpkg' binary package");
-		}
-		return (compareVersionStrings(dpkgInstalledVersion->versionString, "1.16.1") != -1); // >=
-	}
-	else
-	{
-		return config.getBool(optionName);
 	}
 }
 
@@ -2071,7 +2047,7 @@ void PackagesWorker::changeSystem(const shared_ptr< download::Progress >& downlo
 	vector< Changeset > changesets;
 	{
 		GraphAndAttributes gaa;
-		if (!__build_actions_graph(gaa))
+		if (!__build_actions_graph(gaa) && __actions_preview->autoFlagChanges.empty())
 		{
 			_logger->log(Logger::Subsystem::Packages, 1, "nothing to do");
 			return;
@@ -2083,170 +2059,25 @@ void PackagesWorker::changeSystem(const shared_ptr< download::Progress >& downlo
 
 	_logger->log(Logger::Subsystem::Packages, 1, "changing the system");
 
-	// doing or simulating the actions
-	auto dpkgBinary = _config->getPath("dir::bin::dpkg");
-	{
-		auto dpkgOptions = _config->getList("dpkg::options");
-		FORIT(optionIt, dpkgOptions)
-		{
-			dpkgBinary += " ";
-			dpkgBinary += *optionIt;
-		}
-	}
-
-	bool deferTriggers = __defer_triggers(*_config, *_cache);
-
 	__do_dpkg_pre_actions();
-
-	{ // make sure system is trigger-clean
-		_logger->log(Logger::Subsystem::Packages, 2, "running all package triggers");
-
-		auto command = dpkgBinary + " --triggers-only -a";
-		__run_dpkg_command("triggers-only", command, "");
-	}
-
-	auto archivesDirectory = _get_archives_directory();
-	FORIT(changesetIt, changesets)
 	{
-		const Changeset& changeset = *changesetIt;
-
-		if (debugging)
+		for (const Changeset& changeset: changesets)
 		{
-			debug2("started changeset");
-		}
-		/* usually, all downloads are done before any install actions (first
-		   and only changeset) however, if 'cupt::worker::archives-space-limit'
-		   is turned on this is no longer the case, and we will do downloads/installs
-		   by portions ("changesets") */
-		__do_downloads(changeset.downloads, downloadProgress);
+			Dpkg dpkg(this);
 
-		__do_dpkg_pre_packages_actions(changeset.actionGroups);
-
-		FORIT(actionGroupIt, changeset.actionGroups)
-		{
-			/* all actions within one group have, by algorithm,
-			   a) the same action name (so far, configures only)
-			   b) the same package name (linked actions)
-
-			   in both cases, we can choose the action type of the last action
-			   in the subgroup as effective
-			*/
-			InnerAction::Type actionType = actionGroupIt->rbegin()->type;
-
-			string actionName;
-			switch (actionType)
+			if (debugging) debug2("started changeset");
+			__do_downloads(changeset.downloads, downloadProgress);
+			__do_dpkg_pre_packages_actions(changeset.actionGroups);
+			for (const auto& actionGroup: changeset.actionGroups)
 			{
-				case InnerAction::Remove:
-				{
-					const string& packageName = actionGroupIt->rbegin()->version->packageName;
-					actionName = __actions_preview->groups[Action::Purge].count(packageName) ?
-							"purge" : "remove";
-				}
-					break;
-				case InnerAction::Unpack:
-					actionName = "unpack";
-					break;
-				case InnerAction::Configure:
-					if (actionGroupIt->size() >= 2 && (actionGroupIt->rbegin() + 1)->type == InnerAction::Unpack)
-					{
-						actionName = "install"; // [remove+]unpack+configure
-					}
-					else
-					{
-						actionName = "configure";
-					}
-					break;
+				__change_auto_status(actionGroup);
+				dpkg.doActionGroup(actionGroup, *__actions_preview);
 			}
-
-			__change_auto_status(*actionGroupIt);
-
-			{ // dpkg actions
-				auto dpkgCommand = dpkgBinary + " --" + actionName;
-				if (deferTriggers)
-				{
-					dpkgCommand += " --no-triggers";
-				}
-				// add necessary options if requested
-				string requestedDpkgOptions;
-				FORIT(dpkgFlagIt, actionGroupIt->dpkgFlags)
-				{
-					requestedDpkgOptions += string(" ") + *dpkgFlagIt;
-				}
-				dpkgCommand += requestedDpkgOptions;
-
-				/* the workaround for a dpkg bug #558151
-
-				   dpkg performs some IMHO useless checks for programs in PATH
-				   which breaks some upgrades of packages that contains important programs
-
-				   It is possible that this hack is not needed anymore with
-				   better scheduling heuristics of 2.x but we cannot
-				   re-evaluate it with lenny->squeeze (e)glibc upgrade since
-				   new Cupt requires new libgcc which in turn requires new
-				   glibc.
-				*/
-				dpkgCommand += " --force-bad-path";
-
-				FORIT(actionIt, *actionGroupIt)
-				{
-					if (actionIt->type != actionType)
-					{
-						continue; // will be true for non-last linked actions
-					}
-					string actionExpression;
-					if (actionName == "unpack" || actionName == "install")
-					{
-						const shared_ptr< const BinaryVersion > version = actionIt->version;
-						actionExpression = archivesDirectory + '/' + _get_archive_basename(version);
-					}
-					else
-					{
-						actionExpression = actionIt->version->packageName;
-					}
-					dpkgCommand += " ";
-					dpkgCommand += actionExpression;
-				}
-				{ // debug & logs
-					if (debugging)
-					{
-						vector< string > stringifiedActions;
-						FORIT(actionIt, *actionGroupIt)
-						{
-							stringifiedActions.push_back(actionIt->toString());
-						}
-						auto actionsString = join(" & ", stringifiedActions);
-						debug2("do: (%s) %s%s", actionsString, requestedDpkgOptions,
-								actionGroupIt->continued ? " (continued)" : "");
-					}
-					_logger->log(Logger::Subsystem::Packages, 2,
-							__get_dpkg_action_log(*actionGroupIt, actionType, actionName));
-				}
-				_run_external_command(Logger::Subsystem::Packages, dpkgCommand);
-			};
-		}
-		if (deferTriggers)
-		{
-			// triggers were not processed during actions perfomed before, do it now at once
-			_logger->log(Logger::Subsystem::Packages, 2, "running pending triggers");
-
-			string command = dpkgBinary + " --triggers-only --pending";
-			if (debugging)
-			{
-				debug2("running triggers");
-			}
-			__run_dpkg_command("triggers", command, "");
-		}
-
-		if (archivesSpaceLimit)
-		{
-			__clean_downloads(changeset);
-		}
-		if (debugging)
-		{
-			debug2("finished changeset");
+			if (archivesSpaceLimit) __clean_downloads(changeset);
+			if (debugging) debug2("finished changeset");
 		}
+		__do_independent_auto_status_changes();
 	}
-
 	__do_dpkg_post_actions();
 }
 
